/* *****************************************************************************
 * decode/std/addin.c -- decode an add-in file.
 * Copyright (C) 2017 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
 *
 * This file is part of libg1m.
 * libg1m is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License,
 * or (at your option) any later version.
 *
 * libg1m is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with libg1m; if not, see <http://www.gnu.org/licenses/>.
 * ************************************************************************** */
#include <libg1m/internals.h>

/* ************************************************************************** */
/*  Add-in for the fx-9860G (G1A)                                             */
/* ************************************************************************** */
/**
 *	g1m_decode_std_addin:
 *	Decodes a "normal" add-in (after Standard Header).
 *
 *	@arg	h			the handle to make.
 *	@arg	buffer		the buffer to read from.
 *	@return				the error code (0 if ok).
 */

int g1m_decode_std_addin(g1m_handle_t **h, g1m_buffer_t *buffer,
	g1m_standard_header_t *std)
{
	int err = 0;
	(void)std;

	/* get the subheader */
	DREAD(hd, g1m_addin_subheader_s)

	/* make the handle */
	g1m_version_t version; g1m_decode_version(
		(char*)hd.g1m_addin_subheader_version, &version);
	time_t created; g1m_decode_date(
		(char*)hd.g1m_addin_subheader_creation_date, &created);
	size_t content_size =
		be32toh(hd.g1m_addin_subheader_filesize); /* already corrected */
	content_size -= sizeof(g1m_standard_header_t)
		+ sizeof(g1m_addin_subheader_t);
	err = g1m_make_addin(h, g1m_platform_fx, content_size,
		(char*)hd.g1m_addin_subheader_title,
		(char*)hd.g1m_addin_subheader_internal_name, &version, &created);
	if (err) return (err);
	g1m_handle_t *handle = *h;

	/* log info about the subheader */
	log_info("title is '%s'", handle->g1m_handle_title);
	log_info("internal name is '%s'", handle->g1m_handle_intname);
	log_info("estrips count is %" PRIu8, hd.g1m_addin_subheader_estrips_count);
	log_info("version is %02u.%02u",
		handle->g1m_handle_version.g1m_version_major,
		handle->g1m_handle_version.g1m_version_minor);
	log_info("creation date is: %.24s",
		ctime(&handle->g1m_handle_creation_date));

	/* fill icon */
	g1m_decode_picture(handle->g1m_handle_icon_unsel,
		g1m_pictureformat_1bit, hd.g1m_addin_subheader_icon,
		handle->g1m_handle_width, handle->g1m_handle_height);
	g1m_decode_picture(handle->g1m_handle_icon_sel,
		g1m_pictureformat_1bit_r, hd.g1m_addin_subheader_icon,
		handle->g1m_handle_width, handle->g1m_handle_height);

	/* read content */
	GREAD(handle->g1m_handle_content, handle->g1m_handle_size)

	/* no errors */
	return (err);
fail:
	g1m_free(*h); *h = NULL;
	return (err);
}

/* ************************************************************************** */
/*  Add-in for the fx-CP/Classpad (C1A)                                       */
/* ************************************************************************** */
/**
 *	g1m_decode_std_cp_addin:
 *	Decode fx-CP add-in.
 *
 *	@arg	h			the handle to make.
 *	@arg	buffer		the buffer to read from.
 *	@arg	std			the standard header.
 *	@arg	sub			the standard subheader.
 *	@arg	cp			the classpad-specific subheader.
 *	@arg	check		the checksum to feed.
 *	@return				the error code (0 if ok).
 */

int g1m_decode_std_cp_addin(g1m_handle_t **h, g1m_buffer_t *buffer,
	g1m_standard_header_t *std, g1m_standard_subheader_t *sub,
	g1m_standard_classpad_subheader_t *cp, uint32_t *check)
{
	int err = 0;
	(void)std; (void)cp;

	/* read the add-in subheader */
	DREAD(cphd, g1m_classpad_addin_subheader_s)
	*check = g1m_checksum32(&cphd, sizeof(g1m_classpad_addin_subheader_t),
		*check);

	/* make the handle */
	g1m_version_t version; g1m_decode_version(
		(char*)sub->g1m_standard_subheader_version, &version);
	time_t created; g1m_decode_date(
		(char*)sub->g1m_standard_subheader_timestamp, &created);
	size_t content_size = be32toh(sub->g1m_standard_subheader_filesize)
		- 0x1000;
	err = g1m_make_addin(h, g1m_platform_cp, content_size,
		(char*)sub->g1m_standard_subheader_title,
		(char*)sub->g1m_standard_subheader_internal_name,
		&version, &created);
	if (err) return (err);
	g1m_handle_t *handle = *h;

	/* decode pictures */
	g1m_decode_picture(handle->g1m_handle_icon_unsel,
		g1m_pictureformat_1bit_packed, cphd.g1m_classpad_addin_subheader_icon,
		handle->g1m_handle_width, handle->g1m_handle_height);
	g1m_decode_picture(handle->g1m_handle_icon_sel,
		g1m_pictureformat_1bit_packed_r, cphd.g1m_classpad_addin_subheader_icon,
		handle->g1m_handle_width, handle->g1m_handle_height);

	/* log */
	log_info("title is '%s'", handle->g1m_handle_title);
	log_info("internal name is '%s'", handle->g1m_handle_intname);
	log_info("version is %02u.%02u",
		handle->g1m_handle_version.g1m_version_major,
		handle->g1m_handle_version.g1m_version_minor);
	log_info("timestamp is %.24s", ctime(&handle->g1m_handle_creation_date));

	/* get content */
	GREAD(handle->g1m_handle_content, handle->g1m_handle_size)
	*check = g1m_checksum32(handle->g1m_handle_content,
		handle->g1m_handle_size, *check);

	/* no error */
	return (0);
fail:
	g1m_free(*h); *h = NULL;
	return (err);
}

/* ************************************************************************** */
/*  Add-in for the fx-CG/Prizm (G3A)                                          */
/* ************************************************************************** */
/**
 *	g1m_decode_std_cg_addin:
 *	Decode fx-CG add-in.
 *
 *	@arg	h			the handle to make.
 *	@arg	buffer		the buffer to read from.
 *	@arg	std			the standard header.
 *	@arg	sub			the standard subheader.
 *	@arg	prizm		the prizm-specific subheader.
 *	@arg	check		the checksum to feed.
 *	@return				the error code (0 if ok).
 */

int g1m_decode_std_cg_addin(g1m_handle_t **h, g1m_buffer_t *buffer,
	g1m_standard_header_t *std, g1m_standard_subheader_t *sub,
	g1m_standard_prizm_subheader_t *prizm, uint32_t *check)
{
	int err = 0;
	(void)std; (void)prizm;

	/* read the add-in subheader */
	DREAD(cghd, g1m_prizm_addin_subheader_s)
	*check = g1m_checksum32(&cghd, sizeof(g1m_prizm_addin_subheader_t), *check);

	/* make the handle */
	g1m_version_t version; g1m_decode_version(
		(char*)sub->g1m_standard_subheader_version, &version);
	time_t created; g1m_decode_date(
		(char*)sub->g1m_standard_subheader_timestamp, &created);
	size_t content_size = be32toh(sub->g1m_standard_subheader_filesize);
	content_size -= sizeof(g1m_standard_header_t)
		+ sizeof(g1m_standard_subheader_t)
		+ sizeof(g1m_standard_prizm_subheader_t)
		+ sizeof(g1m_prizm_addin_subheader_t)
		+ sizeof(uint32_t);
	log_info("Content size is %" PRIuSIZE, content_size);
	err = g1m_make_addin(h, g1m_platform_cg, content_size,
		(char*)sub->g1m_standard_subheader_internal_name,
		(char*)sub->g1m_standard_subheader_title, &version, &created);
	if (err) return (err);
	g1m_handle_t *handle = *h;

	/* decode pictures */
	g1m_decode_picture(handle->g1m_handle_icon_unsel,
		g1m_pictureformat_16bit,
		cghd.g1m_prizm_addin_subheader_unselected_icon_image,
		handle->g1m_handle_width, handle->g1m_handle_height);
	g1m_decode_picture(handle->g1m_handle_icon_sel,
		g1m_pictureformat_16bit,
		cghd.g1m_prizm_addin_subheader_selected_icon_image,
		handle->g1m_handle_width, handle->g1m_handle_height);

	/* log */
	log_info("title is '%s'", handle->g1m_handle_title);
	log_info("internal name is '%s'", handle->g1m_handle_intname);
	log_info("version is %02u.%02u",
		handle->g1m_handle_version.g1m_version_major,
		handle->g1m_handle_version.g1m_version_minor);
	log_info("timestamp is %.24s", ctime(&handle->g1m_handle_creation_date));

	/* read content */
	GREAD(handle->g1m_handle_content, handle->g1m_handle_size)
	*check = g1m_checksum32(handle->g1m_handle_content,
		handle->g1m_handle_size, *check);

	/* no error */
	return (0);
fail:
	g1m_free(*h); *h = NULL;
	return (err);
}
